{
open Parser


(*** types used for lexer staging ***)

type lexer = Lexing.lexbuf -> token
type stage_spec = string * (token * lexer)
type stage = Required of stage_spec
           | Optional of stage_spec

let stage_name = function Required (name, _) | Optional (name, _) -> name
let spec_name = function (name, _) -> name
let optional = function Required _ -> false | Optional _ -> true

type side = Left | Right of lexer
type state = { mutable side : side; mutable stages : stage list }
            (* left or right side,  stages remaining *)


(*** comment handling ***)

let comment_start_ps = ref []

let enter_comment lexbuf =
    let (start_p, end_p) =
        (Lexing.lexeme_start_p lexbuf, Lexing.lexeme_end_p lexbuf)
    in
    comment_start_ps := (start_p, end_p) :: !comment_start_ps

let leave_comment () =
    match !comment_start_ps with
        | [] -> assert false
        | _::ps -> comment_start_ps := ps


(*** error reporting ***)

(* update the lexbuf's Lexing.position appropriately on a newline *)
let newline lexbuf =
    let curr_p = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <-
        { curr_p with
          Lexing.pos_lnum = curr_p.Lexing.pos_lnum + 1;
          Lexing.pos_bol = curr_p.Lexing.pos_cnum }

(* emit an error based on the lexeme matched *)
let error lexbuf s =
    let pos1 = Lexing.lexeme_start_p lexbuf in
    let pos2 = Lexing.lexeme_end_p lexbuf in
    raise (Syntax.Error (pos1, pos2, s))

(* an invalid character error *)
(* XXX double-check that this is used correctly everywhere! *)
let invalid_char lexbuf lexing =
    error lexbuf ("Invalid character '" ^ Lexing.lexeme lexbuf
                 ^ "' while lexing " ^ lexing)

let lex_category lexbuf cat_str =
    match cat_str with
        | "apps" -> `apps
        | "godi" -> `godi
        | "conf" -> `conf
        | "base" -> `base
        | _ -> error lexbuf ("Invalid category, '" ^ cat_str
                            ^ "' in package name")
}


(*** regular expressions ***)

let digit = ['0'-'9']
(* must not contain dashes, according to pkgsrc documentation *)
(* XXX more documentation?  what *exactly* should we be matching with these? *)
let version_num = digit ['0'-'9' 'a'-'z' 'A'-'Z' '_' '.'] *
(* should these be extended?  should they be complementized? *)
(* let ident = ['a'-'z' 'A'-'Z' '-' '0'-'9' '_' '.'] + *)
let pkgname_cat = ['a'-'z' 'A'-'Z' '0'-'9' '_']+
let pkgname_name = ['a'-'z' 'A'-'Z' '0'-'9' '_' '-']+
let pkgname = (pkgname_cat as category) '-' (pkgname_name as name)
let pathname = ['a'-'z' 'A'-'Z' '-' '0'-'9' '_' '.' '/'] +
let maketarget = ['a'-'z' 'A'-'Z' '0'-'9' '-' '_' '.'] +
let url = ['a'-'z' 'A'-'Z' '-' '0'-'9' '_' '.' '/' ':' '~' '%' '+' '&'] +


(*** lexers ***)

rule lhs = parse
    | [^' ' '\t' '\n']* { Lexing.lexeme lexbuf }

and package = parse
    | pkgname       { PKGNAME (lex_category lexbuf category, name) }
    | pkgname_cat   { error lexbuf ("Expected '-' in package name '"
                                   ^ Lexing.lexeme lexbuf ^ "'") }
    | _             { invalid_char lexbuf "package name" }

and version = parse
    | version_num   { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "version number" }

and revision = parse
    | digit+        { INT (int_of_string (Lexing.lexeme lexbuf)) }
    | _             { invalid_char lexbuf "revision number" }

(* XXX here, and above, make pkgname_cat error message more informative:
 *  - if followed by whitespace, expected a dash.
 *  - otherwise, invalid character
 *)
and depends = parse
    | pkgname       { PKGNAME (lex_category lexbuf category, name) }
    | pkgname_cat   { error lexbuf ("Expected '-' in package name '"
                                   ^ Lexing.lexeme lexbuf ^ "'") }
    | version_num   { STRING (Lexing.lexeme lexbuf) }
    | '('           { LPAREN }
    | ')'           { RPAREN }
    | "=="          { RELOP `eq }
    | ">="          { RELOP `ge }
    | "<="          { RELOP `le }
    | '>'           { RELOP `gt }
    | '<'           { RELOP `lt }
    | ','           { COMMA }
    | _             { invalid_char lexbuf "dependencies" }

and sources = parse
    | url           { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "sources url" }

and unpacksto = parse
    | pathname      { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "Unpacks-To line" }

and homepage = parse
    | url           { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "homepage url" }

and maintainer = parse
    | [^ '\n']*     { STRING (Lexing.lexeme lexbuf) }

and make_target = parse
    | maketarget    { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "make target" }

and description state = parse
    | [^ '\n']* as short '\n'
      ((_* '\n')? as long)
      ".\n" 
                { let s = Lexing.lexeme lexbuf in
                  String.iter (function '\n' -> newline lexbuf | _ -> ()) s;
                  state.side <- Left;
                  STRINGPAIR (short, long) }
    | ""        { error lexbuf
                  "Unterminated description, expected '.' on a line by itself" }

and options = parse
    | "configure"   { OPTION `configure }
    | "bmake"       { OPTION `bmake }
    | "opt"         { OPTION `opt }
    | "htdoc"       { OPTION `htdoc }
    | ','           { COMMA }
    | [^' ' '\t' '\n' ',']*
                    { error lexbuf ("Invalid option '" ^ Lexing.lexeme lexbuf
                                    ^ "', expected one of "
                                    ^ "'configure', 'bmake', 'opt', 'htdoc'") }

and docfiles = parse
    | pathname      { STRING (Lexing.lexeme lexbuf) }
    | ','           { COMMA }
    | _             { invalid_char lexbuf "docfiles" }

and comment = parse
    | "*)"          { leave_comment ();
                      if !comment_start_ps = [] then () else comment lexbuf }
    | "(*"          { enter_comment lexbuf; comment lexbuf }
    | '\n'          { newline lexbuf; comment lexbuf }
    | _             { comment lexbuf }
    | eof           { match !comment_start_ps with
                        | [] -> assert false
                        | (p1, p2)::_ ->
                            raise (Syntax.Error (p1, p2, "Unterminated comment")) }

(* higher-order lexer -- strips whitespace and comments *)
and strip lexer = parse
    | [' ' '\t']+   { strip lexer lexbuf }
    | "(*"          { enter_comment lexbuf; comment lexbuf; strip lexer lexbuf }
    | ""            { lexer lexbuf }

(* like strip, but strips newlines, too.  used by lhs. *)
and strip_nl lexer = parse
    | [' ' '\t']+   { strip_nl lexer lexbuf }
    | "(*"          { enter_comment lexbuf;
                      comment lexbuf;
                      strip_nl lexer lexbuf }
    | '\n'          { newline lexbuf; strip_nl lexer lexbuf }
    | ""            { lexer lexbuf }

(* higher-order lexer -- advances/emits EOL on newline *)
and advance state lexer = parse
    | '\n'          { newline lexbuf; state.side <- Left; EOL }
    | ""            { lexer lexbuf }

{
(* the trailer *)

(*** staging / state management ***)

let start_state () =
    let state = { side = Left;
                  stages = []; }
    in
    (* it is the responsibility of the lexer to advance state.side back to
       Left.  most lexers here make use of `advance state ...', a higher-order
       lexer that advances state.side to Left on newline and emits an EOL.
       in order to advance the state, all lexers must take `state' as a param *)
    state.stages <-
       [
        Required ("Package:", (PACKAGE, advance state package));
        Required ("Version:", (VERSION, advance state version));
        Required ("Revision:", (REVISION, advance state revision));
        Required ("Depends:", (DEPENDS, advance state depends));
        Optional ("Build-Depends:", (BUILD_DEPENDS, advance state depends));
        Required ("Sources:", (SOURCES, advance state sources));
        Optional ("Unpacks-To:", (UNPACKSTO, advance state unpacksto));
        Optional ("Bytecode-Target:", (BC_TARGET, advance state make_target));
        Optional ("Opt-Target:", (OPT_TARGET, advance state make_target));
        Required ("Homepage:", (HOMEPAGE, advance state homepage));
        Required ("Maintainer:", (MAINTAINER, advance state maintainer));
        Optional ("Options:", (OPTIONS, advance state options));
        Optional ("Docfiles:", (DOCFILES, advance state docfiles));
        Required ("Description:", (DESCRIPTION, description state));
       ];
    state

let get_state =
    let states = ref [] in
    fun lexbuf ->
        try
            List.assq lexbuf !states
        with
            Not_found ->
                let state = start_state () in
                states := (lexbuf, state) :: !states;
                state

let ordered = ref true (* option *)
(* things that depend on ordered:
 *  - possible_next_stages
 *  - stages_after_finishing
 *)

(* possible_next_stages : stage list -> stage_spec list *)
let possible_next_stages stages =
    if !ordered then
    (* if ordered, possible next stages are any optionals plus next required *)
        (* TODO: make tail recursive, maybe *)
        let rec take_through_required = function
            | Required spec :: rest -> [spec]
            | Optional spec :: rest -> spec :: take_through_required rest
            | [] -> []
        in
        take_through_required stages
    else
    (* if not ordered, possible next stages are any of them, optional or not *)
        List.map (function Required spec | Optional spec -> spec) stages

let stages_after_finishing name stages =
    if !ordered then
        (* if ordered, we remove everything up to and incl. named stage *)
        let rec remove_through name = function
            | stage :: stages when stage_name stage = name -> stages
            | _ :: stages -> remove_through name stages
            | [] -> assert false
              (* if we've finished `name', it must exist in the list *)
        in
        remove_through name stages
    else
        (* if not ordered, we remove only the named stage *)
        let rec remove name = function
            | stage :: stages when stage_name stage = name -> stages
            | stage :: stages -> stage :: remove name stages
            | [] -> assert false
        in
        remove name stages

let advance_to_rhs state lexeme lexbuf =
    let possibilities = possible_next_stages state.stages in
    try
        let (token, lexer) = List.assoc lexeme possibilities in
        state.side <- Right lexer;
        state.stages <- stages_after_finishing lexeme state.stages;
        token
    with
        Not_found ->
            let string_of_possibilities = function
                | [] -> "EOF"
                | [spec] -> "'" ^ spec_name spec ^ "'"
                | specs -> "one of '"
                          ^ String.concat "', '" (List.map spec_name specs)
                          ^ "'"
            in
            let stringify_lexeme = function
                | "" -> "EOF"
                | s -> "'" ^ s ^ "'"
            in
            (* XXX: if all possibilities came from Optionals,
               we should also say we could have expected EOF *)
            error lexbuf ("Expected " ^ string_of_possibilities possibilities
                         ^ ", got " ^ stringify_lexeme lexeme)


(*** the main "lexer" ***)

let token lexbuf =
    let state = get_state lexbuf in
    match state.side with
        | Left ->
            (* lhs always returns *something* -- either
               "Field:", "<garbage>", or "" (for eof) *)
            let lexeme = strip_nl lhs lexbuf in
            if lexeme = "" (* eof *) && List.for_all optional state.stages then
            (* we've encountered eof and all remaining stages are optional *)
                EOF
            else
            (* we must try to advance to the next stage based on the input *)
                let token = advance_to_rhs state lexeme lexbuf in
                token
                (* advances state to (Right (something from state.stages))
                   and removes that thing from stages, or errors *)
        | Right lexer ->
            strip lexer lexbuf

let open_lex filename =
    let chan = open_in filename in
    let lexbuf = Lexing.from_channel chan in
    lexbuf.Lexing.lex_curr_p <- { lexbuf.Lexing.lex_curr_p
                                  with Lexing.pos_fname = filename };
    lexbuf
}
